import json
from datetime import datetime
from typing import Any, Dict, Optional, Union

from termcolor import colored

from mloggers._log_levels import LogLevel
from mloggers.logger import Logger


class ConsoleLogger(Logger):
    """Logs to the console (i.e., standard I/O)."""

    def log(
        self,
        message: Union[str, Dict[str, Any]],
        level: Optional[Union[LogLevel, str]] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super(ConsoleLogger, self).log(message, level, *args, **kwargs)

        time = "[" + datetime.now().strftime("%H:%M:%S") + "]"

        if level is None:
            level_str = ""
            level_clr = ""
        else:
            # Create the level string to be printed
            if isinstance(level, LogLevel):
                level_str = "[" + level.name + "] "
            else:
                level_str = "[" + str(level).upper() + "] "

            # Color the level string
            if isinstance(level, LogLevel):
                level_clr = colored(level_str, level.value)
            else:
                level_clr = colored(level_str, "green")

        # The first level of the dictionary is printed as a multiline
        # indented message.
        # The rest of the levels are printed as a single line
        # pretifyed depending on the type of the value.
        if isinstance(message, dict):
            first = True
            for key, value in message.items():
                if not first:
                    time = " " * len(time)
                    level_clr = " " * len(level_str)

                if isinstance(value, float):
                    print(f"{level_clr}{time} {key}: {value:.5f}")
                elif isinstance(value, Union[dict, list]):
                    value = json.dumps(value, indent=4)
                    print(f"{level_clr}{time} {key}: {value}")
                elif value is None:  # Used for headers, titles, etc.
                    print(f"{level_clr}{time} {key}")
                else:
                    print(f"{level_clr}{time} {key}: {value}")

                first = False

        elif hasattr(message, "__str__") and callable(getattr(message, "__str__")):
            print(f"{level_clr}{time} {str(message)}")



from enum import Enum

__logger_level_members = {
    "WARN": "yellow",
    "ERROR": "red",
    "DEBUG": "magenta",
    "INFO": "cyan",
}

LogLevel = Enum("LogLevel", __logger_level_members)


def register_level(level: str, color: str):
    """
    Register a customized logger level, which will then be available as a member of `LogLevel`,
    where its `name` is the argument `level` and its `value` is the argument `color`.
    Note: the name of the level will be uppercased.

    ### Parameters
    ----------
    `level` -> the level name to register.
    `color` -> the color to use for the level.
    """

    level = level.upper()
    __logger_level_members.update({level: color})
    LogLevel[level] = color



from typing import Any, Dict, Optional, Union

import wandb
from omegaconf import DictConfig
from termcolor import colored

from mloggers._log_levels import LogLevel
from mloggers.logger import Logger


class WandbLogger(Logger):
    """Logs to Weights & Biases."""

    def __init__(
        self, project: str, group: str, experiment: str, config: Optional[DictConfig]
    ):
        """
        Initializes a Weights & Biases logger.

        ### Parameters
        ----------
        `project` -> the name of the project to log to.
        `group` -> the name of the group to log to.
        `experiment` -> the name of the experiment to log to.
        [optional] `config` -> the configuration of the experiment.
        """

        if config is not None:
            config = vars(config)
        wandb.init(project=project, group=group, name=experiment, config=config)

    def log(
        self,
        message: Union[str, Dict[str, Any]],
        level: Optional[Union[LogLevel, str]] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super(WandbLogger, self).log(message, level, *args, **kwargs)

        if isinstance(message, dict):
            log = message
        else:
            if level is not None:
                level_str = (
                    level.name if isinstance(level, LogLevel) else str(level).upper()
                )
                log = {level_str: message}
            else:
                log = {"message": message}

        try:
            wandb.log(log)
        except Exception as e:
            print(f'{colored("[ERROR]", "red")} Error while logging to wandb: {e}')

    def __del__(self):
        """Finishes the connection with W&B."""

        wandb.finish()
